package sbc.mgclus.algorithm;

import java.util.HashSet;
import java.util.PriorityQueue;
import java.util.Set;

import org.apache.log4j.Logger;
import org.jgrapht.WeightedGraph;
import org.jgrapht.graph.DefaultWeightedEdge;

import sbc.mgclus.algorithm.scoring.Scoring;
import sbc.mgclus.data.*;

public class ClusteringKL {
	private int mergesPerIteration;// = 100;
	private double delta;// = 0.0;
	private WeightedGraph<Cluster, DefaultWeightedEdge> graph;
	private PriorityQueue<ClusterPair> clustersToMerge;
	private Scoring scoring;
	private Integer newClusterID;
	private Logger logger=Logger.getLogger(ClusteringKL.class);
	
	public ClusteringKL(WeightedGraph<Cluster, DefaultWeightedEdge> g, Scoring s){
		this(g, s, 0.000001);
	}
	
	public ClusteringKL(WeightedGraph<Cluster, DefaultWeightedEdge> g, Scoring s, double delta){
		graph = g;
		scoring = s;
		//calculate the node degree for each cluster/vertex
		calcNodeDegree();
		newClusterID = (graph.vertexSet()).size()+1;
		this.delta = delta;
		this.mergesPerIteration = 100;
		logger.debug(String.format("merges per iteration: %s", mergesPerIteration));
	}
	
	public void performClusteringKL(){
		//get the initial list of cluster pairs
		clustersToMerge = scoring.getInitialClusterPairs();
		//as long as their is cluster pairs with a positive merge gain -> merge them
		int iteration = 1;
		int mergesPI = 100;
		logger.info("start clustering...\n");
		while (clustersToMerge.size() > 0){
			//merge top 100 cluster pairs (only if none of them was merged before)
			HashSet<Cluster> newClusters = new HashSet<Cluster>();
			int count = 0;
			int attempts = 0;
			while (count < mergesPI){
				ClusterPair clusterPair = clustersToMerge.poll();
				if (clusterPair == null)
					break;
				else{
					Cluster c1 = clusterPair.getCluster1();
					Cluster c2 = clusterPair.getCluster2();
					if (graph.containsVertex(c1) && graph.containsVertex(c2)){
						Cluster mergedCluster = mergeClusterPair(c1, c2);
						newClusters.add(mergedCluster);
						//debugging
						//logger.debug(String.format("merge-> cluster pair: c%s, c%s, iteration: %d, merge: %d, merge gain: %f, c%s", c1.getClusterID(), c2.getClusterID(), iteration, count, clusterPair.getMergeGain(), mergedCluster.getClusterID()));
						//logger.debug("\t");
						//for (Vertex v : mergedCluster.getMembers()){
						//	logger.debug(v.getName()+"\t");
						//	//logger.debug(String.format("%s"), v.getName());
						//}
						//logger.debug("\n");
					}
					//else{
					//	logger.debug(String.format("skip-> cluster pair: c%s, c%s, iteration: %d, merge: %d", c1.getClusterID(), c2.getClusterID(), iteration, count));
					//}
					count++;
				}
				attempts++;
			}
			//filter out cluster pairs that are not any longer in the graph
			PriorityQueue<ClusterPair> remainingClusterPairs = new PriorityQueue<ClusterPair>();
			for (ClusterPair cp : clustersToMerge){
				Cluster c1 = cp.getCluster1();
				Cluster c2 = cp.getCluster2();
				if (graph.containsVertex(c1) && graph.containsVertex(c2)){
					remainingClusterPairs.add(cp);
				}
			}
			clustersToMerge = remainingClusterPairs;
			//add edges of merged clusters to clusters pairs that needs to be merged
			//stop clustering if the number of new clusters is < 2
			if (newClusters.size() > 1){
				//debugging
				logger.debug(String.format("iteration: %d: %d merges performed...allowed merges in this iteration %d...attempts %d",iteration, newClusters.size(), mergesPI, attempts));
				HashSet<DefaultWeightedEdge> processedEdges = new HashSet<DefaultWeightedEdge>();
				for (Cluster newCluster : newClusters){
					for (DefaultWeightedEdge edge : graph.edgesOf(newCluster)){
						if (! processedEdges.contains(edge)){
							Cluster c1 = graph.getEdgeSource(edge);
							Cluster c2 = graph.getEdgeTarget(edge);
							double mergeGain = scoring.calcMergeGain(c1, c2);
							//debugging
							//logger.debug(String.format("merge gains for: c%s, c%s, %f", c1.getClusterID(), c2.getClusterID(), mergeGain));
							if (mergeGain > delta){
								ClusterPair cPair = new ClusterPair(c1, c2, mergeGain);
								clustersToMerge.add(cPair);
							}
							processedEdges.add(edge);
						}
					}
				}
			}
			iteration ++;
		}
	}
	
	private Cluster mergeClusterPair(Cluster c1, Cluster c2){
		//get edges connected to c1 and c2
		Set<DefaultWeightedEdge> edgesOfC1 = graph.edgesOf(c1);
		Set<DefaultWeightedEdge> edgesOfC2 = graph.edgesOf(c2);
		HashSet<Cluster> neighbors = new HashSet<Cluster>();
		//find all neighbors of c1 and c2 ... this can be optimized but might be ok for testing
		for (DefaultWeightedEdge edgeC1 : edgesOfC1){
			Cluster s1 = graph.getEdgeSource(edgeC1);
			Cluster t1 = graph.getEdgeTarget(edgeC1);
			if (! s1.equals(c1))
				neighbors.add(s1);
			else
				neighbors.add(t1);
		}
		for (DefaultWeightedEdge edgeC2 : edgesOfC2){
			Cluster s2 = graph.getEdgeSource(edgeC2);
			Cluster t2 = graph.getEdgeTarget(edgeC2);
			if (! s2.equals(c2))
				neighbors.add(s2);
			else
				neighbors.add(t2);
		}
		//initialize a new "merged" cluster
		Cluster mergedCluster = new Cluster(newClusterID);
		newClusterID += 1;
		mergedCluster.setDegree(c1.getDegree()+c2.getDegree());
		graph.addVertex(mergedCluster);
		mergedCluster.addMemberList(c1.getMembers());
		mergedCluster.addMemberList(c2.getMembers());
		mergedCluster.setSeedCount(c1.getSeedCount()+c2.getSeedCount());
		//add a new edge between all old neighbors and the new merged cluster
		for (Cluster neighbor : neighbors){
			double newEdgeWeight = 0;
			if (graph.containsEdge(c1, neighbor))
				newEdgeWeight += graph.getEdgeWeight(graph.getEdge(c1, neighbor));
			if (graph.containsEdge(c2, neighbor))
				newEdgeWeight += graph.getEdgeWeight(graph.getEdge(c2, neighbor));
			graph.addEdge(mergedCluster, neighbor);
			graph.setEdgeWeight(graph.getEdge(mergedCluster, neighbor), newEdgeWeight);
		}
		//recalculate the efficiency of the merged cluster
		double mergedEfficiency = graph.getEdgeWeight(graph.getEdge(c1, c2))+c1.getClusterEfficiency()+c2.getClusterEfficiency();
		mergedEfficiency += scoring.calcCNScore(c1, c2);
		mergedCluster.setClusterEfficiency(mergedEfficiency);
		//remove the old cluster
		graph.removeVertex(c1);
		graph.removeVertex(c2);
		
		return mergedCluster;
	}
	
	private void calcNodeDegree(){
		//for each cluster set the cluster degree equal to the sum of all edge weights
		for (Cluster c : graph.vertexSet()){
			double nodeDegree = 0;
			for (DefaultWeightedEdge e : graph.edgesOf(c)){
				nodeDegree += graph.getEdgeWeight(e);
			}
			c.setDegree(nodeDegree);
			//logger.debug(String.format("cluster c%s, degree %f", c.getClusterID(), nodeDegree));
		}
	}
	
}
